import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.spatial.distance import pdist, squareform
from scipy.cluster.hierarchy import dendrogram, linkage

# Load the dataset
file_path = 'iris.csv'  
df = pd.read_csv(file_path)

# Use only the numerical columns for clustering
df_numeric = df.drop(columns=["species"])
dataset_np = df_numeric.to_numpy()

def plot_dendrogram_fixed(data, method):
    """
    Plots a dendrogram using the hierarchical clustering linkage matrix.
    
    Parameters:
    - data: ndarray, the dataset for clustering.
    - method: str, the linkage method to compute the clustering (e.g., 'single', 'complete', 'average').
    """
    # Plot dendrograms for all linkage methods in a single figure
fig, axes = plt.subplots(1, 3, figsize=(18, 6), sharey=True)

for ax, method in zip(axes, ['single', 'complete', 'average']):
    ax.set_title(f'Dendrogram ({method.capitalize()} Linkage)')
    ax.set_xlabel('Sample index')
    ax.set_ylabel('Distance' if method == 'single' else "")
    
    # Compute the linkage matrix
    linkage_matrix = linkage(dataset_np, method=method)
    
    # Plot the dendrogram
    dendrogram(linkage_matrix, ax=ax, color_threshold=0.7 * max(linkage_matrix[:, 2]))
    
plt.tight_layout()
plt.show()

    
def diana(data: np.ndarray, max_clusters: int = 3, method='single'):
    """
    Implements the DIANA (Divisive Analysis) clustering algorithm.
    
    Parameters:
    - data: ndarray, the dataset for clustering.
    - max_clusters: int, the maximum number of clusters to create.
    - method: str, the linkage method to compute the clustering (e.g., 'single', 'complete', 'average').
    
    Returns:
    - clusters: list of ndarray, containing cluster indices.
    """
    distance_matrix = squareform(pdist(data))
    clusters = [np.arange(data.shape[0], dtype=np.uint16)]
    counter = 0
    
    while len(clusters) < max_clusters:
        counter += 1
        print(f"Iteration: {counter}")
        
        cluster_sizes = [True if c.shape[0] <= 1 else False for c in clusters]
        if all(cluster_sizes):
            break
        
        can_divide = True
        
        # Find diameter
        diameter_list = []
        for cluster in clusters:
            diameter_indv_list = []
            if len(cluster) == 1:
                diameter_indv_list.append(-99)
            for point_a in cluster:
                for point_b in cluster:
                    if point_a == point_b:
                        continue
                    diameter_indv_list.append(distance_matrix[point_a][point_b])
            diameter_list.append(max(diameter_indv_list) if diameter_indv_list else -1)
        
        chosen_cluster_idx = np.argmax(diameter_list)
        C_i = clusters.pop(chosen_cluster_idx)
        C_j = np.array([], dtype=np.uint16)
        
        while can_divide:
            # Calculate dissimilarity
            dis_list = []
            for point_a in C_i:
                dis_total_a = 0
                if len(C_i) > 1:
                    dis_factor_a = 1 / (len(C_i) - 1)
                    for point_b in C_i:
                        if point_a == point_b:
                            continue
                        dis_total_a += distance_matrix[point_a][point_b]
                    dis_total_a *= dis_factor_a
                
                dis_total_b = 0
                if len(C_j) > 0:
                    dis_factor_b = 1 / len(C_j)
                    for point_b in C_j:
                        if point_a == point_b:
                            continue
                        dis_total_b += distance_matrix[point_a][point_b]
                    dis_total_b *= dis_factor_b
                
                dissimilarity = dis_total_a - dis_total_b
                dis_list.append(dissimilarity)
            
            chosen_point_idx = np.argmax(dis_list)
            if dis_list[chosen_point_idx] < 0:
                can_divide = False
                break
            
            C_j = np.append(C_j, C_i[chosen_point_idx])
            C_i = np.delete(C_i, chosen_point_idx)
        
        clusters.append(C_j)
        clusters.append(C_i)
    return clusters

# Perform DIANA clustering with different linkage methods
linkage_methods = ['single', 'complete', 'average']
max_clusters = 3

for method in linkage_methods:
    print(f"\nUsing {method.capitalize()} Linkage:")
    final_clusters = diana(dataset_np, max_clusters=max_clusters, method=method)
    
    # Plot dendrogram with the specified linkage method
    plot_dendrogram_fixed(dataset_np, method=method)

    # Print final clusters with IDs
    print("\nFinal Clusters (IDs):")
    for i, cluster in enumerate(final_clusters):
        print(f"Cluster {i + 1}: {cluster}")